/*  
    DeSpot - Conditional Temporal Despotting Filter for Avisynth 2.5
    Main file
    
    Version 1.0  December 30, 2003
    with "sign", "fitluma", "blur" , "tsmooth" parameters added by Fizick     

Copyright (C)2003 Alexander G. Balakhnin aka Fizick.
bag@hotmail.ru         http://bag.hotmail.ru
under the GNU General Public Licence version 2.

This filter is based on  Conditional Temporal Median Filter for Avisynth 2.5
Version 0.93, September 27, 2003
Copyright (C) 2003 Kevin Atkinson (kevin.med@atkinson.dhs.org) under the GNU GPL version 2.
http://kevin.atkinson.dhs.org/temporal_median/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "avisynth_c.h"
#include "despot.hpp"

#ifdef DEBUG

#include <iostream>
#include <ext/stdio_filebuf.h>
#include <windows.h>
#include <io.h>
#include <fcntl.h>
using std::cerr;
using std::cout;
using std::ostream;
using std::endl;

struct Log : public ostream {
  __gnu_cxx::stdio_filebuf<char> fbuf;
  int get_fd() {
    AllocConsole();
    HANDLE h = GetStdHandle( STD_OUTPUT_HANDLE );
    return _open_osfhandle( (long) h, _O_TEXT );}
  Log() : ostream(&fbuf), fbuf(get_fd(), std::ios::out, false, 128) {}
} clog;

#else //DEBUG

struct Log {
  template <typename T>
  Log & operator<<(const T &) {return *this;}
} clog;
#define endl 0

#endif //DEBUG

struct WorkingData {
  int num;
  byte * noise;
  byte * motion;
  Border * borders;
  void init(size_t);
  WorkingData() : num(-1), noise(0), motion(0), borders(0) {}
  ~WorkingData();
};

struct Frame {
  int num;
  WorkingData * data;
  AVS_VideoFrame * frame;
  ConstRow y;
  byte * noise;
  byte * motion;
  Borders borders;
  Frame() : num(-1), data(0), frame(0), noise(0), motion(0) {}
  ~Frame();
};

struct Buffer
{
  AVS_Clip * child;
  WorkingData data_cache[2];
  Frame frame_cache[4];
  int num_frames;
  
  Frame * get_frame(int n)
    {if (n < 0 || n >= num_frames) return 0;
     Frame * f = frame_cache + n % 4;             // % is remainder of integer division
     if (f->num != n) ready_frame(n, f);
     return f;}
  
  void alloc_noise(Frame * f)
    {if (!f->data) ready_data(f);
     f->noise = f->data->noise;}

  void alloc_motion(Frame * f)
    {if (!f->data) ready_data(f);
     f->motion = f->data->motion;}
  
  void alloc_borders(Frame * f)
    {if (!f->data) ready_data(f);
     f->borders.data = f->data->borders;
     f->borders.clear();}

  REGPARM void ready_frame(int f, Frame * f);
  REGPARM void ready_data(Frame * f);
  REGPARM void free_frame(Frame * f);
  REGPARM void free_data(WorkingData *);
};

struct Filter : public Parms 
{
  AVS_FilterInfo * inf;
  Buffer buf;
  
  REGPARM void find_noise(int fn);
  REGPARM void find_motion(int fn);
  REGPARM void find_motion_simple(int fn);
  REGPARM AVS_VideoFrame * exec(int fn);
  
  static AVS_VideoFrame * get_frame(AVS_FilterInfo *, int n);
  static void free_filter(AVS_FilterInfo *);
};


/////////////////////////////////////////////////////////////////////
//
// Filter
//

Exec * exec_f[8] = { remove_outliers, mark_outliers, map_outliers, 0,
                     cond_median, map_outliers, map_outliers, 0     };

void Filter::find_noise(int fn)
{
  clog << "Find Noise: " << fn << endl;
  Frame * p = buf.get_frame(fn-1);
  Frame * c = buf.get_frame(fn);
  buf.alloc_noise(c); buf.alloc_borders(c);
  Frame * n = buf.get_frame(fn+1);
  if (p && n) {
    ::find_outliers(p->y, c->y, n->y, c->noise, *this);
    ::find_sizes(c->borders, c->noise, *this);
  }  
  ::mark_noise(c->borders, c->noise, *this);
}

void Filter::find_motion(int fn)
{
  clog << "Find Motion: " << fn << endl;
  Frame * p = buf.get_frame(fn-1);
  if (!p->noise) find_noise(fn-1);
  Frame * c = buf.get_frame(fn);
  if (!c->noise) find_noise(fn);
  buf.alloc_motion(c);
  ::mark_motion(p->y, p->noise, c->y, c->noise, c->motion, *this);
  ::motion_denoise(c->motion, *this);
}

void Filter::find_motion_simple(int fn)
{
  clog << "Find Motion Simple: " << fn << endl;
  Frame * p = buf.get_frame(fn-1);
  Frame * c = buf.get_frame(fn);
  buf.alloc_motion(c);
  ::find_motion(p->y, c->y, c->motion, *this);
  ::motion_denoise(c->motion, *this);
}

AVS_VideoFrame * Filter::exec(int fn)
{
  clog << "Exec: " << fn << endl;
  Frame * p = buf.get_frame(fn-1);
  Frame * c = buf.get_frame(fn);
  Frame * n = buf.get_frame(fn+1);
  if (p && n) {
    AVS_VideoFrame * fout = avs_new_video_frame(inf->env, &inf->vi);
    if (mode == M_SEARCH) {
      if (!c->noise) find_noise(fn);
      if (!c->motion) find_motion(fn);
      if (!n->motion) find_motion(fn+1);
    } else {
      if (!c->motion) find_motion_simple(fn);
      if (!n->motion) find_motion_simple(fn+1);
    }
    MutableRow fout_y;
    fout_y.data = avs_get_write_ptr(fout);
    fout_y.pitch = avs_get_pitch(fout);
    clog << "Render " << (mode | show) << ": " << fn << endl;
    (*exec_f[mode|show])(p->y, c->y, c->noise, c->motion, n->y, n->motion,
                         fout_y, *this);
    if (show != S_MAP || show_chroma) {
      avs_bit_blt(inf->env, avs_get_write_ptr_p(fout, AVS_PLANAR_U),
                            avs_get_pitch_p(fout, AVS_PLANAR_U),
                            avs_get_read_ptr_p(c->frame, AVS_PLANAR_U),
                            avs_get_pitch_p(c->frame, AVS_PLANAR_U),
                            avs_get_row_size_p(c->frame, AVS_PLANAR_U),
                            avs_get_height_p(c->frame, AVS_PLANAR_U));
      avs_bit_blt(inf->env, avs_get_write_ptr_p(fout, AVS_PLANAR_V),
                            avs_get_pitch_p(fout, AVS_PLANAR_V),
                            avs_get_read_ptr_p(c->frame, AVS_PLANAR_V),
                            avs_get_pitch_p(c->frame, AVS_PLANAR_V),
                            avs_get_row_size_p(c->frame, AVS_PLANAR_V),
                            avs_get_height_p(c->frame, AVS_PLANAR_V));
    } else {
      memset(avs_get_write_ptr_p(fout, AVS_PLANAR_U), 0x80,
             avs_get_pitch_p(fout, AVS_PLANAR_U)*avs_get_height_p(c->frame, AVS_PLANAR_U));
      memset(avs_get_write_ptr_p(fout, AVS_PLANAR_V), 0x80,
             avs_get_pitch_p(fout, AVS_PLANAR_V)*avs_get_height_p(c->frame, AVS_PLANAR_V));
    }
    return fout;
  } else {
    return avs_copy_frame(c->frame);
  }
}

AVS_VideoFrame * Filter::get_frame(AVS_FilterInfo * fi, int n)
{
  clog << "Get Frame: " << n << endl;
  Filter * f = (Filter *)fi->user_data;
  return f->exec(n);
}

void Filter::free_filter(AVS_FilterInfo * fi)
{
  Filter * f = (Filter *)fi->user_data;
  delete f;
  fi->user_data = 0;
}

#define ERR(str) do {ret = avs_new_value_error(str); goto error;} while (0)
#define SET_I(what, min, max) do {\
    tmp = avs_array_elt(args, i++);\
    if (avs_defined(tmp)) {\
      int v = avs_as_int(tmp);\
      if (v < min || v > max)\
        ERR(#what " must be in the range " #min "-" #max ".");\
      f->what = v;}} while (0)
#define SET_BYTE(what) SET_I(what, 0, 255)
#define SET_UINT(what) SET_I(what, 0, 32500)
//Added by Fizick:
#define SET_SIGN(what) SET_I(what, -2, 2)

extern "C"
AVS_Value apply_filter(AVS_ScriptEnvironment * env, 
                       AVS_Value args, void * user_data)
{
  AVS_Value ret = avs_void;
  int mask = (int)user_data;
  Filter * f = new Filter;
  int i = 0;
  AVS_Clip * clip = avs_new_c_filter(env, &f->inf, avs_array_elt(args, i++), true);
  if (!avs_is_yv12(&f->inf->vi)) ERR("Video must be YV12");
  f->buf.child = f->inf->child;
  f->buf.num_frames = f->inf->vi.num_frames;
  f->inf->get_frame = Filter::get_frame;
  f->inf->free_filter = Filter::free_filter;
  f->inf->user_data = f;
  f->height = f->inf->vi.height;
  f->width  = f->inf->vi.width;
  f->y_next = avs_is_field_based(&f->inf->vi) ? 2 : 1;
  if (f->width % 8 != 0) ERR("Video Width Must be a multiple of 8");
  f->size   = f->height * f->width;
  f->buf.data_cache[0].init(f->size);
  f->buf.data_cache[1].init(f->size);
  f->mode = mask & ~3;
  f->show = mask &  3;
  AVS_Value tmp;
  // common parms
  SET_BYTE(mthres);
  SET_UINT(mwidth);
  SET_UINT(mheight);
  f->mp = f->mwidth*f->mheight/3;  // New default value of mp, added by Fizick
  SET_BYTE(mp);
  tmp = avs_array_elt(args, i++);
  if (avs_defined(tmp)) f->y_next = avs_as_bool(tmp) ? 2 : 1;
  if (f->mode == M_SEARCH) 
  {
    SET_BYTE(p1);
    SET_BYTE(p2);
    SET_UINT(pwidth);
    SET_UINT(pheight);
    SET_SIGN(sign);                                        //Added by Fizick
    tmp = avs_array_elt(args, i++);                        //Added by Fizick
    if (avs_defined(tmp)) f->fitluma = avs_as_bool(tmp);   //Added by Fizick
    SET_BYTE(blur);                                        //Added by Fizick
    SET_BYTE(tsmooth);                                     //Added by Fizick
  }
  if (f->show == S_MARK) {
    SET_BYTE(mark_v);
  } else if (f->show == S_MAP) {
    tmp = avs_array_elt(args, i++);
    if (avs_defined(tmp)) f->show_chroma = avs_as_bool(tmp);
  }
  ret = avs_new_value_clip(clip);
error:
  avs_release_clip(clip); clip = 0;
  return ret;
}

#define common_parms  "c[mthres]i[mwidth]i[mheight]i[mp]i[interlaced]b"
//define denoise_parms "[p1]i[p2]i[pwidth]i[pheight]i"
//Changed by Fizick:
#define denoise_parms "[p1]i[p2]i[pwidth]i[pheight]i[sign]i[fitluma]b[blur]i[tsmooth]i"
extern "C"
const char * avisynth_c_plugin_init(AVS_ScriptEnvironment* env)
{
  avs_add_function(env, "DeSpot", 
                   common_parms denoise_parms,
                   apply_filter, (void *)(M_SEARCH));
  avs_add_function(env, "DeSpotMark",
                   common_parms denoise_parms "[mark_v]i",
                   apply_filter, (void *)(M_SEARCH | S_MARK));
  avs_add_function(env, "DeSpotMap",
                   common_parms denoise_parms "[show_chroma]b",
                   apply_filter, (void *)(M_SEARCH | S_MAP));
  avs_add_function(env, "DeSpotMedian", 
                   common_parms,
                   apply_filter, (void *)(M_MEDIAN));
  avs_add_function(env, "DeSpotMedianMap", 
                   common_parms "[show_chroma]b",
                   apply_filter, (void *)(M_MEDIAN | S_MAP));
  return "";
}

/////////////////////////////////////////////////////////////////////
//
// Frame
//

Frame::~Frame()
{
  if (frame) avs_release_video_frame(frame);
}

//
// WorkingData
//

void WorkingData::init(size_t s)
{
  noise   = (byte *)malloc(s);
  motion  = (byte *)malloc(s);
  borders = (Border *)malloc(border_size * (s/2+1));
}

WorkingData::~WorkingData()
{
  if (noise)   free(noise);
  if (motion)  free(motion);
  if (borders) free(borders);
}

//
// Buffer
//

void Buffer::ready_frame(int n, Frame * f)
{
  free_frame(f);
  f->num = n;
  clog << "Getting Frame: " << n << "\n";
  f->frame = avs_get_frame(child, n);         
  f->y.data  = avs_get_read_ptr(f->frame);
  f->y.pitch = avs_get_pitch(f->frame);
}

void Buffer::ready_data(Frame * f)
{
  WorkingData * d = data_cache + f->num % 2;
  free_data(d);
  d->num = f->num;
  f->data = d;
}

void Buffer::free_frame(Frame * f)
{
  if (f->data) free_data(f->data);
  clog << "Freeing Frame: " << f->num << "\n";
  if (f->frame) avs_release_video_frame(f->frame);
  f->num = -1;
  f->frame = 0;
  f->y.data = 0;
}

void Buffer::free_data(WorkingData * d)
{
  Frame * f = frame_cache + d->num % 4;
  if (f->num == d->num) {
    f->data = 0;
    f->noise = 0;
    f->motion = 0;
    f->borders.data = 0;
    f->borders.clear();
  }
  d->num = -1;
}

